#include "OgreVisibilityOctreeSceneManager.h"
#include "OgreVisibilityOptionsManager.h"
#include <OgreMath.h>
#include <OgreIteratorWrappers.h>
#include <OgreRenderSystem.h>
#include <OgreCamera.h>
#include <OgreLogManager.h>
#include <OgreStringConverter.h>
#include <OgreEntity.h>
#include <OgreSubEntity.h>
#include <OgreMaterialManager.h>
#include <OgreIteratorWrappers.h>
#include "VspBspTree.h"
#include "Containers.h"
#include "ViewCellsManager.h"
#include <OgreConfigFile.h>
#include "OgreTypeConverter.h"
#include "OgreMeshInstance.h"
#include "common.h"

// normal terrain rendering
const static NORMAL_RENDER_HACK = false;

namespace Ogre {

//-----------------------------------------------------------------------
VisibilityOctreeSceneManager::VisibilityOctreeSceneManager(const String& name, 
	GtpVisibility::VisibilityManager *visManager)
:
OctreeSceneManager(name),
mVisibilityManager(visManager), 
mShowVisualization(false),
mRenderNodesForViz(false),
mRenderNodesContentForViz(false),
mVisualizeCulledNodes(false),
mLeavePassesInQueue(0),
mDelayRenderTransparents(true),
mUseDepthPass(false),
mIsDepthPassPhase(false),
mUseItemBuffer(false),
mIsItemBufferPhase(false),
mCurrentEntityId(1),
mEnableDepthWrite(true),
mSkipTransparents(false),
mRenderTransparentsForItemBuffer(true),
mExecuteVertexProgramForAllPasses(true),
mIsHierarchicalCulling(false),
mViewCellsLoaded(false),
mUseViewCells(false),
mUseVisibilityFilter(false),
mCurrentViewCell(NULL),
mElementaryViewCell(NULL)
{
	mHierarchyInterface = new OctreeHierarchyInterface(this, mDestRenderSystem);
		
	if (0)
	{
		mDisplayNodes = true;
		mShowBoundingBoxes = true;
		mShowBoxes = true;
	}

	// TODO: set maxdepth to reasonable value
	mMaxDepth = 50;

	//loadVisibilityConfig("GtpVisibility.cfg");
	//LoadViewCells("");
	//mViewCellsLoaded = true;
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::InitDepthPass()
{
	MaterialPtr depthMat = MaterialManager::getSingleton().getByName("Visibility/DepthPass");

	if (depthMat.isNull())
    {
		depthMat = MaterialManager::getSingleton().create(
                "Visibility/DepthPass",
                ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);

		mDepthPass = depthMat->getTechnique(0)->getPass(0);
		mDepthPass->setColourWriteEnabled(false);
		mDepthPass->setDepthWriteEnabled(true);
		mDepthPass->setLightingEnabled(false);
	}
	else
	{
		mDepthPass = depthMat->getTechnique(0)->getPass(0);
	}
}
//-----------------------------------------------------------------------
VisibilityOctreeSceneManager::~VisibilityOctreeSceneManager()
{
	OGRE_DELETE(mHierarchyInterface);
	CLEAR_CONTAINER(mObjects);
	OGRE_DELETE(mCurrentViewCell);
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::InitItemBufferPass()
{
	MaterialPtr itemBufferMat = MaterialManager::getSingleton().
		getByName("Visibility/ItemBufferPass");

	if (itemBufferMat.isNull())
    {
		// Init
		itemBufferMat = MaterialManager::getSingleton().create("Visibility/ItemBufferPass",
                ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);

		mItemBufferPass = itemBufferMat->getTechnique(0)->getPass(0);
		mItemBufferPass->setColourWriteEnabled(true);
		mItemBufferPass->setDepthWriteEnabled(true);
		mItemBufferPass->setLightingEnabled(true);
		//mItemBufferPass->setLightingEnabled(false);
	}
	else
	{
		mItemBufferPass = itemBufferMat->getTechnique(0)->getPass(0);
	}
	//mItemBufferPass->setAmbient(1, 1, 0);
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::PrepareVisualization(Camera *cam)
{
	// add player camera for visualization purpose
	try 
	{
		Camera *c;
		if ((c = getCamera("PlayerCam")) != NULL)
		{
			getRenderQueue()->addRenderable(c);
		}   
    }
    catch(...)
    {
        // ignore
    }
	// add bounding boxes of rendered objects
	if (1)
	for (BoxList::iterator it = mBoxes.begin(); it != mBoxes.end(); ++it)
{
		getRenderQueue()->addRenderable(*it);
	}
	
	if (mRenderNodesForViz || mRenderNodesContentForViz)
	{
		// HACK: change node material so it is better suited for visualization
		MaterialPtr nodeMat = MaterialManager::getSingleton().getByName("Core/NodeMaterial");
		nodeMat->setAmbient(1, 1, 0);
		nodeMat->setLightingEnabled(true);
		nodeMat->getTechnique(0)->getPass(0)->removeAllTextureUnitStates();

		for (NodeList::iterator it = mVisible.begin(); it != mVisible.end(); ++it)
		{
			if (mRenderNodesForViz)
			{
				// render the leaf nodes
				if ((*it)->numAttachedObjects() && 
					!(*it)->numChildren() && 
					((*it)->getAttachedObject(0)->getMovableType() == "Entity") &&
					(*it)->getAttachedObject(0)->isVisible())
				{
				getRenderQueue()->addRenderable((*it));
			}
			}
			// add renderables itself
			if (mRenderNodesContentForViz) 
			{
				(*it)->_addToRenderQueue(cam, getRenderQueue(), false);
			}
		}
	}
}
//-----------------------------------------------------------------------
const Pass *VisibilityOctreeSceneManager::_setPass(Pass* pass)
{
	if (NORMAL_RENDER_HACK)
	{
		return SceneManager::_setPass(pass);
	}

	// TODO: setting vertex program is not efficient
	//Pass *usedPass = ((mIsDepthPassPhase && !pass->hasVertexProgram()) ? mDepthPass : pass); 
	
	// set depth fill pass if we currently do not make an aabb occlusion query
	const bool useDepthPass = 
		(mIsDepthPassPhase && !mHierarchyInterface->IsBoundingBoxQuery());

	Pass *usedPass = useDepthPass ? mDepthPass : pass;

	IlluminationRenderStage savedStage = mIlluminationStage; 
	
	// set illumination stage to NONE so no shadow material is used 
	// for depth pass or for occlusion query
	if (mIsDepthPassPhase || mHierarchyInterface->IsBoundingBoxQuery())
	{
		mIlluminationStage = IRS_NONE;
	}

	// --- set vertex program of current pass in order to set correct depth
	if (mExecuteVertexProgramForAllPasses && 
		mIsDepthPassPhase && 
		pass->hasVertexProgram())
    {
		// add vertex program of current pass to depth pass
		mDepthPass->setVertexProgram(pass->getVertexProgramName());

		if (mDepthPass->hasVertexProgram())
		{
			const GpuProgramPtr& prg = mDepthPass->getVertexProgram();
			// Load this program if not done already
			if (!prg->isLoaded())
				prg->load();
			// Copy params
			mDepthPass->setVertexProgramParameters(pass->getVertexProgramParameters());
		}
	}
	else if (mDepthPass->hasVertexProgram()) // reset vertex program
	{
		mDepthPass->setVertexProgram("");
	}

	// save old depth write: needed for item buffer
	const bool IsDepthWrite = usedPass->getDepthWriteEnabled();

	// global option which enables / disables depth writes
	if (!mEnableDepthWrite)
	{
		usedPass->setDepthWriteEnabled(false);
	}


	//-- set actual pass here
	const Pass *result = SceneManager::_setPass(usedPass);


	// reset depth write
	if (!mEnableDepthWrite)
	{
		usedPass->setDepthWriteEnabled(IsDepthWrite);
	}

	// reset illumination stage
	mIlluminationStage = savedStage;

	return result;
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::_findVisibleObjects(Camera* cam, bool onlyShadowCasters)
{
	if (NORMAL_RENDER_HACK)
	{
		OctreeSceneManager::_findVisibleObjects(cam, onlyShadowCasters);
		return;
	}
	

	//-- show visible scene nodes and octree bounding boxes from last frame
	if (mShowVisualization)
    {
		PrepareVisualization(cam);
	}
	else 
	{	
		// for hierarchical culling, we interleave identification 
		// and rendering of objects in _renderVisibibleObjects

		// for the shadow pass we use only standard rendering
		// because of low occlusion
		if (mShadowTechnique == SHADOWTYPE_TEXTURE_MODULATIVE && 
			mIlluminationStage == IRS_RENDER_TO_TEXTURE)
		{
			OctreeSceneManager::_findVisibleObjects(cam, onlyShadowCasters);
		}

		// only shadow casters will be rendered in shadow texture pass
		if (0) mHierarchyInterface->SetOnlyShadowCasters(onlyShadowCasters);


		//-- apply view cell pvs
		updatePvs(cam);
	}
	
	
	// -- delete lists stored for visualization
	mVisible.clear();
	mBoxes.clear();
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::_renderVisibleObjects()
{
	if (NORMAL_RENDER_HACK)
	{
		OctreeSceneManager::_renderVisibleObjects();
	
		return;
	}

	InitDepthPass(); 	  // create material for depth pass
	InitItemBufferPass(); // create material for item buffer pass

	// save ambient light to reset later
	ColourValue savedAmbient = mAmbientLight;

	//-- apply standard rendering for some modes (e.g., visualization, shadow pass)

	if (mShowVisualization ||
	   (mShadowTechnique == SHADOWTYPE_TEXTURE_MODULATIVE && 
	    mIlluminationStage == IRS_RENDER_TO_TEXTURE))
	{
		IlluminationRenderStage savedStage = mIlluminationStage; 
	
		if (mShowVisualization)
		{
			// disable illumination stage to prevent rendering shadows
			mIlluminationStage = IRS_NONE;
		}

		// standard rendering for shadow maps because of performance
		OctreeSceneManager::_renderVisibleObjects();

		mIlluminationStage = savedStage;
	}
	else //-- the hierarchical culling algorithm
	{	
		// this is also called in TerrainSceneManager: really necessary?
		//mDestRenderSystem -> setLightingEnabled(false);

		// don't render backgrounds for item buffer
		if (mUseItemBuffer)
		{
			clearSpecialCaseRenderQueues();
			getRenderQueue()->clear();
		} 	

		//-- hierarchical culling
		// the objects of different layers (e.g., background, scene, 
		// overlay) must be identified and rendered one after another

		//-- render all early skies
		clearSpecialCaseRenderQueues();
		addSpecialCaseRenderQueue(RENDER_QUEUE_BACKGROUND);
		addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_EARLY);
		setSpecialCaseRenderQueueMode(SceneManager::SCRQM_INCLUDE);

		OctreeSceneManager::_renderVisibleObjects();
		/////////////////////////////////////////////////

#ifdef GTP_VISIBILITY_MODIFIED_OGRE
		// delete previously rendered content
		_deleteRenderedQueueGroups();
#endif

		//-- prepare queue for visible objects (i.e., all but overlay and skies late)
		clearSpecialCaseRenderQueues();
		addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_LATE);
		addSpecialCaseRenderQueue(RENDER_QUEUE_OVERLAY);
		
		// exclude this queues from hierarchical rendering
		setSpecialCaseRenderQueueMode(SceneManager::SCRQM_EXCLUDE);
	

		// set all necessary parameters for 
		// hierarchical visibility culling and rendering
		InitVisibilityCulling(mCameraInProgress);
	
	
		/** 
		* the hierarchical culling algorithm
		* for depth pass: we just find objects and update depth buffer
		* for "delayed" rendering: we render some passes afterwards
		* e.g., transparents, because they need front-to-back sorting
		**/
	
		mVisibilityManager->ApplyVisibilityCulling();
	
		// delete remaining renderables from queue:
		// all which are not in mLeavePassesInQueue)
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
		_deleteRenderedQueueGroups(mLeavePassesInQueue);
#endif

		//-- reset parameters
		mIsDepthPassPhase = false;
		mIsItemBufferPhase = false;
		mSkipTransparents = false;
		mIsHierarchicalCulling = false;

		mLeavePassesInQueue = 0;

#if 1	
		// add visible nodes found by the visibility culling algorithm
		if (mUseDepthPass)
		{
			NodeList::const_iterator it, it_end = mVisible.end();

			//getRenderQueue()->clear();
			for (it = mVisible.begin(); it != it_end; ++ it)
			{
				(*it)->_addToRenderQueue(mCameraInProgress, getRenderQueue(), false);
			}
		}
#endif	
		//-- now we can render all remaining queue objects
		//-- used for depth pass, transparents, overlay
		clearSpecialCaseRenderQueues();
	
		//-- we render all remaining queue objects
		// used for depth pass, transparents, overlay 
		clearSpecialCaseRenderQueues();
		OctreeSceneManager::_renderVisibleObjects();
	
	}   // hierarchical culling

	// reset ambient light
	setAmbientLight(savedAmbient);

	getRenderQueue()->clear(); // finally clear render queue
	if (0) OGRE_DELETE(mRenderQueue); // HACK: should rather only be cleared ...

	if (0) WriteLog(); // write out stats
}

//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::_updateSceneGraph(Camera* cam)
{
	if (NORMAL_RENDER_HACK)
	{
		OctreeSceneManager::_updateSceneGraph(cam);
		return;
	}

	mVisibilityManager->GetCullingManager()->SetHierarchyInterface(mHierarchyInterface);
	mHierarchyInterface->SetRenderSystem(mDestRenderSystem);

	OctreeSceneManager::_updateSceneGraph(cam);
}
//-----------------------------------------------------------------------
bool VisibilityOctreeSceneManager::setOption(const String & key, const void * val)
{
	if (key == "UseDepthPass")
	{
		mUseDepthPass = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "PrepareVisualization")
	{
		mShowVisualization = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesForViz")
	{
		mRenderNodesForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesContentForViz")
	{
		mRenderNodesContentForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyBoxEnabled")
	{
		mSkyBoxEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyPlaneEnabled")
	{
		mSkyPlaneEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyDomeEnabled")
	{
		mSkyDomeEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "VisualizeCulledNodes")
	{
		mVisualizeCulledNodes = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "DelayRenderTransparents")
	{
		mDelayRenderTransparents = (*static_cast<const bool *>(val));
		return true;
	}

	if (key == "DepthWrite")
	{
		mEnableDepthWrite = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "UseItemBuffer")
	{
		mUseItemBuffer = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "ExecuteVertexProgramForAllPasses")
	{ 
		mExecuteVertexProgramForAllPasses  = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderTransparentsForItemBuffer")
	{ 
		mRenderTransparentsForItemBuffer  = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "NodeVizScale")
	{ 
		OctreeNode::setVizScale(*static_cast<const float *>(val));
		return true;
	}
	if (key == "LoadViewCells")
	{ 
		if (!mViewCellsLoaded)
		{
			LoadViewCells(static_cast<const char *>(val));
			mViewCellsLoaded = true;	
		}

		return true;
	}
	if (key == "UseViewCells")
	{ 
		mUseViewCells = *static_cast<const bool *>(val);

		// reset view cell
		OGRE_DELETE(mCurrentViewCell);
		if (mUseViewCells)
			mCurrentViewCell =  mViewCellsManager->GenerateViewCell();
		mElementaryViewCell = NULL;
		// if using view cells, all objects are set to false initially
		SetObjectsVisible(!mUseViewCells);
		
		return true;
	}
	if (key == "UseVisibilityFilter")
	{
		mUseVisibilityFilter = *static_cast<const bool *>(val);
		// set null =>recomputation of the pvs
        mElementaryViewCell = NULL;
		return true;
	}
	
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		setOption(key, val) || OctreeSceneManager::setOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityOctreeSceneManager::getOption(const String & key, void *val)
{
	if (key == "NumHierarchyNodes")
	{
		* static_cast<unsigned int *>(val) = (unsigned int)mNumOctants;
		return true;
	}

	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOption(key, val) && OctreeSceneManager::getOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityOctreeSceneManager::getOptionValues(const String & key, StringVector  &refValueList)
{
	return OctreeSceneManager::getOptionValues( key, refValueList );
}
//-----------------------------------------------------------------------
bool VisibilityOctreeSceneManager::getOptionKeys(StringVector & refKeys)
{
	return  VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOptionKeys (refKeys) || OctreeSceneManager::getOptionKeys(refKeys);
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::setVisibilityManager(GtpVisibility::VisibilityManager *visManager)
{
    mVisibilityManager = visManager;
}
//-----------------------------------------------------------------------
GtpVisibility::VisibilityManager *VisibilityOctreeSceneManager::getVisibilityManager()
{
	return mVisibilityManager;
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::WriteLog()
{
	std::stringstream d;

	d << "Depth pass: " << StringConverter::toString(mUseDepthPass) << ", "
	  << "Delay transparents: " << StringConverter::toString(mDelayRenderTransparents) << ", "
	  << "Use optimization: " << StringConverter::toString(mHierarchyInterface->GetTestGeometryForVisibleLeaves()) << ", "
	  << "Algorithm type: " << mVisibilityManager->GetCullingManagerType() << ", "
	  << "Hierarchy nodes: " << mNumOctants << ", " 
	  << "Traversed nodes: " << mHierarchyInterface->GetNumTraversedNodes() << ", "
	  << "Rendered nodes: " << mHierarchyInterface->GetNumRenderedNodes() << ", "
	  << "Query culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumQueryCulledNodes() << ", "
	  << "Frustum culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumFrustumCulledNodes() << ", "
      << "Queries issued: " << mVisibilityManager->GetCullingManager()->GetNumQueriesIssued() << ", "
	  << "Found objects: " << (int)mVisible.size() << "\n";

	LogManager::getSingleton().logMessage(d.str());
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::renderObjects(
			const QueuedRenderableCollection& objs, 
			QueuedRenderableCollection::OrganisationMode om,
            bool doLightIteration, const LightList* manualLightList)
{
	// for correct rendering, transparents must be rendered after hierarchical culling
	// => do nothing
	if (NORMAL_RENDER_HACK || !mSkipTransparents)
	{
		OctreeSceneManager::renderObjects(objs, om, doLightIteration, manualLightList);
	}
}
//-----------------------------------------------------------------------
bool VisibilityOctreeSceneManager::validatePassForRendering(Pass* pass)
{
	if (NORMAL_RENDER_HACK)
	{
		return SceneManager::validatePassForRendering(pass);
	}

	// skip all but first pass if we are doing the depth pass
	if ((mIsDepthPassPhase || mIsItemBufferPhase) && (pass->getIndex() > 0))
	{
		return false;
	}
	// all but first pass
	/*else if ((!mIsDepthPassPhase || mIsItemBufferPhase) && (pass->getIndex() != 0))
	{
		return false;
	}*/

	return SceneManager::validatePassForRendering(pass);
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::_renderQueueGroupObjects(
	RenderQueueGroup* pGroup, QueuedRenderableCollection::OrganisationMode om)
{
	if (NORMAL_RENDER_HACK || !mIsItemBufferPhase)
	{
		OctreeSceneManager::_renderQueueGroupObjects(pGroup, om);
		return;
	}
#ifdef ITEM_BUFFER
	//-- renders item buffer

    // Iterate through priorities
    RenderQueueGroup::PriorityMapIterator groupIt = pGroup->getIterator();

	while (groupIt.hasMoreElements())
    {
		RenderItemBuffer(groupIt.getNext());
	}
#endif // ITEM_BUFFER
}
#ifdef ITEM_BUFFER
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::RenderItemBuffer(RenderPriorityGroup* pGroup)
{
	// Do solids
	RenderPriorityGroup::SolidRenderablePassMap solidObjs = pGroup->_getSolidPasses();

	// ----- SOLIDS LOOP -----
	RenderPriorityGroup::SolidRenderablePassMap::const_iterator ipass, ipassend;
	ipassend = solidObjs.end();

	for (ipass = solidObjs.begin(); ipass != ipassend; ++ipass)
	{
		// Fast bypass if this group is now empty
		if (ipass->second->empty()) 
			continue;

		// Render only first pass of renderable as false color
		if (ipass->first->getIndex() > 0)
			continue;

		RenderPriorityGroup::RenderableList* rendList = ipass->second;
		
		RenderPriorityGroup::RenderableList::const_iterator irend, irendend;
		irendend = rendList->end();
			
		for (irend = rendList->begin(); irend != irendend; ++irend)
		{
			if (0)
			{
			std::stringstream d; d << "itembuffer, pass name: " << 
				ipass->first->getParent()->getParent()->getName();
				
			LogManager::getSingleton().logMessage(d.str());
			}
			
			RenderSingleObjectForItemBuffer(*irend, ipass->first);
		}
	}

	//-- TRANSPARENT LOOP: must be handled differently from solids

	// transparents are treated either as solids or completely discarded
	if (mRenderTransparentsForItemBuffer)
	{
		RenderPriorityGroup::TransparentRenderablePassList transpObjs = 
			pGroup->_getTransparentPasses();
		RenderPriorityGroup::TransparentRenderablePassList::const_iterator 
			itrans, itransend;

		itransend = transpObjs.end();
		for (itrans = transpObjs.begin(); itrans != itransend; ++itrans)
		{
			// like for solids, render only first pass
			if (itrans->pass->getIndex() == 0)
			{	
				RenderSingleObjectForItemBuffer(itrans->renderable, itrans->pass);
			}
		} 
	}
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::RenderSingleObjectForItemBuffer(Renderable *rend, Pass *pass)
{
	static LightList nullLightList;
	
	int col[4];
	
	// -- create color code out of object id
	col[0] = (rend->getId() >> 16) & 255;
	col[1] = (rend->getId() >> 8) & 255;
	col[2] = rend->getId() & 255;
//	col[3] = 255;

	//mDestRenderSystem->setColour(col[0], col[1], col[2], col[3]);
    
	mItemBufferPass->setAmbient(ColourValue(col[0] / 255.0f,
										    col[1] / 255.0f, 
											col[2] / 255.0f, 1));

	// set vertex program of current pass
	if (mExecuteVertexProgramForAllPasses && pass->hasVertexProgram())
	{
		mItemBufferPass->setVertexProgram(pass->getVertexProgramName());

		if (mItemBufferPass->hasVertexProgram())
		{
			const GpuProgramPtr& prg = mItemBufferPass->getVertexProgram();
			// Load this program if not done already
			if (!prg->isLoaded())
				prg->load();
			// Copy params
			mItemBufferPass->setVertexProgramParameters(pass->getVertexProgramParameters());
		}
	}
	else if (mItemBufferPass->hasVertexProgram())
	{
		mItemBufferPass->setVertexProgram("");
	}

	Pass *usedPass = setPass(mItemBufferPass); 


	// render a single object, this will set up auto params if required
	renderSingleObject(rend, usedPass, false, &nullLightList);
}
#endif // ITEM_BUFFER
//-----------------------------------------------------------------------
GtpVisibility::VisibilityManager *VisibilityOctreeSceneManager::GetVisibilityManager()
{
	return mVisibilityManager;
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::InitVisibilityCulling(Camera *cam)
{
	// reset culling manager stats
	mVisibilityManager->GetCullingManager()->InitFrame(mVisualizeCulledNodes);

	// set depth pass flag before rendering
	mIsDepthPassPhase = mUseDepthPass;

	mIsHierarchicalCulling = true; // during hierarchical culling

	// item buffer needs full ambient lighting to use item colors as unique id
	if (mUseItemBuffer) 
	{
		mIsItemBufferPhase = true;
		setAmbientLight(ColourValue(1,1,1,1));
	}


	// set passes which are stored in render queue 
	// for rendering AFTER hierarchical culling, i.e., passes which need 
	// a special rendering order
	
	mLeavePassesInQueue = 0;

	if (!mUseDepthPass && !mUseItemBuffer)
	{
		if (mShadowTechnique == SHADOWTYPE_STENCIL_ADDITIVE)
		{
			// TODO: remove this pass because it should be processed during hierarchical culling
			//mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_NOSHADOW;

			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_DECAL;
			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_DIFFUSE_SPECULAR;
			mLeavePassesInQueue |= RenderPriorityGroup::TRANSPARENT_PASSES;

			// just render ambient passes
			/*** msz: no more IRS_AMBIENT, see OgreSceneManager.h ***/
			// mIlluminationStage = IRS_AMBIENT;
			getRenderQueue()->setSplitPassesByLightingType(true);
		}
	
		if (mShadowTechnique == SHADOWTYPE_STENCIL_MODULATIVE)
		{
			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_NOSHADOW;
			mLeavePassesInQueue |= RenderPriorityGroup::TRANSPARENT_PASSES;
		}
	
		// transparents should be rendered after hierarchical culling to 
		// provide front-to-back ordering
		if (mDelayRenderTransparents)
		{
			mLeavePassesInQueue |= RenderPriorityGroup::TRANSPARENT_PASSES;
		}
	}

	// skip rendering transparents during the hierarchical culling
	// (because they will be rendered afterwards)
	mSkipTransparents = 
		(mIsDepthPassPhase || (mLeavePassesInQueue & RenderPriorityGroup::TRANSPARENT_PASSES));

	// -- initialise interface for rendering traversal of the hierarchy
	mHierarchyInterface->SetHierarchyRoot(mOctree);
	
	// possible two cameras (one for culling, one for rendering)
	mHierarchyInterface->InitTraversal(mCameraInProgress, 
							mCullCamera ? getCamera("CullCamera") : NULL,
							mLeavePassesInQueue);
		
	//std::stringstream d; d << "leave passes in queue: " << mLeavePassesInQueue;
	//LogManager::getSingleton().logMessage(d.str());
}
//-----------------------------------------------------------------------
OctreeHierarchyInterface *VisibilityOctreeSceneManager::GetHierarchyInterface()
{
	return mHierarchyInterface;
}
//-----------------------------------------------------------------------
Entity* VisibilityOctreeSceneManager::createEntity(const String& entityName, 
													const String& meshName)
{
	Entity *ent = SceneManager::createEntity(entityName, meshName);

	for (int i = 0; i < (int)ent->getNumSubEntities(); ++i)
	{
		ent->getSubEntity(i)->setId(mCurrentEntityId);
	}

	// increase counter of entity id values
	++ mCurrentEntityId;

	return ent;
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::renderAdditiveStencilShadowedQueueGroupObjects(
	RenderQueueGroup* pGroup, QueuedRenderableCollection::OrganisationMode om)
{
	// only render solid passes during hierarchical culling
	if (mIsHierarchicalCulling)
	{
		RenderQueueGroup::PriorityMapIterator groupIt = pGroup->getIterator();
	    LightList lightList;

		while (groupIt.hasMoreElements())
		{
			RenderPriorityGroup* pPriorityGrp = groupIt.getNext();

			// Sort the queue first
			pPriorityGrp->sort(mCameraInProgress);

			// Clear light list
			lightList.clear();

			// Render all the ambient passes first, no light iteration, no lights
			/*** msz: no more IRS_AMBIENT, see OgreSceneManager.h ***/
			// mIlluminationStage = IRS_AMBIENT;

			OctreeSceneManager::renderObjects(pPriorityGrp->getSolidsBasic(), om, false, &lightList);
			// Also render any objects which have receive shadows disabled
			OctreeSceneManager::renderObjects(pPriorityGrp->getSolidsNoShadowReceive(), om, true);		
#if 0		
			std::stringstream d; 
			d << " solid size: " << (int)pPriorityGrp->_getSolidPasses().size()
				<< " solid no shadow size: " << (int)pPriorityGrp->_getSolidPassesNoShadow().size()
				<< "difspec size: " << (int)pPriorityGrp->_getSolidPassesDiffuseSpecular().size()
				<< " decal size: " << (int)pPriorityGrp->_getSolidPassesDecal().size();
			LogManager::getSingleton().logMessage(d.str());
#endif
		}
	}
	else // render the rest of the passes
	{
		OctreeSceneManager::renderAdditiveStencilShadowedQueueGroupObjects(pGroup, om);
	}
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::renderModulativeStencilShadowedQueueGroupObjects(
	RenderQueueGroup* pGroup, QueuedRenderableCollection::OrganisationMode om)
{
   if (mIsHierarchicalCulling)
   {
	   // Iterate through priorities
	   RenderQueueGroup::PriorityMapIterator groupIt = pGroup->getIterator();

	   while (groupIt.hasMoreElements())
	   {
		   RenderPriorityGroup* pPriorityGrp = groupIt.getNext();

		   // Sort the queue first
		   pPriorityGrp->sort(mCameraInProgress);

		   // Do (shadowable) solids
		   OctreeSceneManager::renderObjects(pPriorityGrp->getSolidsBasic(), om, true);
	   }
   }
   else
   {
	   SceneManager::renderModulativeStencilShadowedQueueGroupObjects(pGroup, om);
   }
}
//-------------------------------------------------------------------------
void VisibilityOctreeSceneManager::SetObjectsVisible(const bool visible)
{
	GtpVisibilityPreprocessor::ObjectContainer::iterator it, it_end = mObjects.end();

	for (it = mObjects.begin(); it != it_end; ++ it)
	{
		OgreMeshInstance *omi = static_cast<OgreMeshInstance *>(*it);
		Entity *ent = omi->GetMesh();
		ent->setVisible(visible);
	}
}
//-------------------------------------------------------------------------
void VisibilityOctreeSceneManager::loadVisibilityConfig(const String& filename)
{
	// TODO matt
	// Set up the options
	ConfigFile config;
	String val;

	config.load(filename);

	val = config.getSetting("Algorithm");

    if (!val.empty())
	{
		 VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).setOption("Algorithm", val.c_str());
	}

	val = config.getSetting("UseDepthPass");

	if (!val.empty())
	{
		 setOption("UseDepthPass", val.c_str());
	}
}
//-------------------------------------------------------------------------
inline static AxisAlignedBox EnlargeBox(const AxisAlignedBox &box)
{
	const float eps = 1e-3f;
	const Vector3 veps(eps, eps, eps);
	Vector3 max = box.getMaximum();
	Vector3 min = box.getMinimum();

	return AxisAlignedBox(min - veps, max + veps);
}
//-----------------------------------------------------------------------
Entity *VisibilityOctreeSceneManager::FindCorrespondingObject(const AxisAlignedBox &box)
{
	list<SceneNode *> sceneNodeList;
	AxisAlignedBox mybox = EnlargeBox(box);
	//AxisAlignedBox dummy(Vector3(-50000, -50000, -50000), Vector3(50000, 50000, 50000));
	
	// get intersecting scene nodes
	findNodesIn(mybox, sceneNodeList, NULL);
	

	list<SceneNode *>::const_iterator sit, sit_end = sceneNodeList.end();

	float overlap = 0;//GtpVisibilityPreprocessor::Limits::Small;

	Entity *bestFittingObj = NULL;
	float bestFit = overlap;

	// perfect fit threshold
	const float thresh = 1.0 - GtpVisibilityPreprocessor::Limits::Small;

	
	// find the bbox which is closest to the current bbox
	for (sit = sceneNodeList.begin(); sit != sceneNodeList.end(); ++ sit)
	{
		SceneNode *sn = *sit;
		SceneNode::ObjectIterator oit = sn->getAttachedObjectIterator();

        while (oit.hasMoreElements())
		{
			MovableObject *mo = oit.getNext();

			// we are only interested in scene entities
			if (mo->getMovableType() != "Entity")
			{
				continue;
			}
			 
			const AxisAlignedBox bbox = EnlargeBox(mo->getWorldBoundingBox());
					

			// compute measure how much aabbs overlap
			overlap = RatioOfOverlap(OgreTypeConverter::ConvertFromOgre(mybox),
									 OgreTypeConverter::ConvertFromOgre(bbox));
	
			 if (overlap > bestFit)
			 {
				 bestFit = overlap;
			
				 bestFittingObj = static_cast<Entity *>(mo);

				 // perfect fit => object found, eraly exit
				 if (overlap >= thresh)
					 return bestFittingObj;				
			 }
		}
	}

	if (0)
	{
		std::stringstream d; 
		if (bestFittingObj)
			d << "best fit: " << bestFit;     
		else
			d << "warning, objects do not fit\n" << box;
	
		Ogre::LogManager::getSingleton().logMessage(d.str());
	}

	return bestFittingObj;
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::LoadViewCells(string filename)
{
	// the objects are set to invisible a prioriy
	SetObjectsVisible(false);

	// identify the corresponding Ogre meshes using the bounding boxes
	IdentifyObjects(mObjects);

	// load the view cells assigning the found objects to the pvss
	mViewCellsManager = 
		GtpVisibilityPreprocessor::ViewCellsManager::LoadViewCells(filename, &mObjects);

	std::stringstream d;
	d << "view cells loaded" << endl;
	Ogre::LogManager::getSingleton().logMessage(d.str());
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::IdentifyObjects(GtpVisibilityPreprocessor::ObjectContainer &objects)
{
	const string bboxesFilename = "boxes.out";

	GtpVisibilityPreprocessor::IndexedBoundingBoxContainer iboxes;
	mViewCellsManager->LoadBoundingBoxes(bboxesFilename, iboxes);

	GtpVisibilityPreprocessor::IndexedBoundingBoxContainer::
		const_iterator iit, iit_end = iboxes.end();

	for (iit = iboxes.begin(); iit != iit_end; ++ iit)
	{
		  const GtpVisibilityPreprocessor::AxisAlignedBox3 box = (*iit).second;
		  const AxisAlignedBox currentBox = OgreTypeConverter::ConvertToOgre(box);
		 
		  Entity *ent = FindCorrespondingObject(currentBox);

		  // create new mesh instance
		  OgreMeshInstance *omi = new OgreMeshInstance(ent);
		  omi->SetId((*iit).first);
		  objects.push_back(omi);
	}
}
//-------------------------------------------------------------------------
void VisibilityOctreeSceneManager::applyViewCellPvs(GtpVisibilityPreprocessor::ViewCell *vc, 
													const bool load)
{	// NOTE: should not happen, rather apply view cell representing unbounded space then
	if (!vc) 
	{	
		// set everything visible for savety
		SetObjectsVisible(true);

		return;
	}

	GtpVisibilityPreprocessor::ObjectPvsMap::const_iterator oit,
			oit_end = vc->GetPvs().mEntries.end();

	//-- PVS of view cell
	for (oit = vc->GetPvs().mEntries.begin(); oit != oit_end; ++ oit)
	{
		if (!(*oit).first) continue;

		OgreMeshInstance *omi = dynamic_cast<OgreMeshInstance*>((*oit).first);
		omi->GetMesh()->setVisible(load);
		//GtpVisibilityPreprocessor::Debug << "here45 " << omi->GetId() << endl;
	}
}
//-------------------------------------------------------------------------
void VisibilityOctreeSceneManager::updatePvs(Camera *cam)
{
	if (mViewCellsLoaded && mUseViewCells)
	{
		const GtpVisibilityPreprocessor::Vector3 viewPoint = 
			OgreTypeConverter::ConvertFromOgre(cam->getDerivedPosition());

		GtpVisibilityPreprocessor::ViewCell *newElementary = 
			mViewCellsManager->GetViewCell(viewPoint);

		// elementary view cell did not change => apply same pvs
		if (mElementaryViewCell == newElementary)
			return;

		mElementaryViewCell = newElementary;

		//-- unload old pvs
		applyViewCellPvs(mCurrentViewCell, false);


		//-- the new view cell
		
		GtpVisibilityPreprocessor::ViewCell *viewCell;

		
		if (mUseVisibilityFilter)
		{
			//-- compute new filtered cell
			GtpVisibilityPreprocessor::PrVs prvs;
			mViewCellsManager->GetPrVS(viewPoint, prvs);
			viewCell = prvs.mViewCell;
		}
		else
		{
			viewCell = newElementary;
		}
			
		//-- load new pvs
		applyViewCellPvs(viewCell, true);

		// store pvs
		mCurrentViewCell->SetPvs(viewCell->GetPvs());

		// delete merge tree of filtered view cell
		if (mUseVisibilityFilter)
			mViewCellsManager->DeleteLocalMergeTree(viewCell);
	}
}



//-----------------------------------------------------------------------
const String VisibilityOctreeSceneManagerFactory::FACTORY_TYPE_NAME = "VisibilityOctreeSceneManager";
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManagerFactory::initMetaData(void) const
{
	mMetaData.typeName = FACTORY_TYPE_NAME;
	mMetaData.description = "Scene manager organising the scene on the basis of an octree with advanced occlusion culling (TM).";
	mMetaData.sceneTypeMask = 0xFFFF; // support all types
	mMetaData.worldGeometrySupported = false;
}
//-----------------------------------------------------------------------
SceneManager* VisibilityOctreeSceneManagerFactory::createInstance(
	const String& instanceName)
{
	return new VisibilityOctreeSceneManager(instanceName, visManager);
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManagerFactory::destroyInstance(SceneManager* instance)
{
	delete instance;
}

}  // namespace Ogre